// SOARA Elmer Saturday May 2018, Temperature, Pressure, Humitity, Altitude example.
// This example uses the Adafruit TFT and BME sensor libraries
// 73,
// Ed, WA6ED

// Remove comment // to enable debug options
// add or remove // comment to select BME280 or BME680 but not both!
#define BME280
//#define BME680

//#define DEBUG

/***************************************************
  This is a library for the Adafruit 1.8" SPI display.

This library works with the Adafruit 1.8" TFT Breakout w/SD card
  ----> http://www.adafruit.com/products/358
The 1.8" TFT shield
  ----> https://www.adafruit.com/product/802
The 1.44" TFT breakout
  ----> https://www.adafruit.com/product/2088
as well as Adafruit raw 1.8" TFT display
  ----> http://www.adafruit.com/products/618

  Check out the links above for our tutorials and wiring diagrams
  These displays use SPI to communicate, 4 or 5 pins are required to
  interface (RST is optional)
  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/

#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_ST7735.h> // Hardware-specific library
#include <Wire.h>
#include <SPI.h>
#include <Adafruit_Sensor.h>
#include <EEPROM.h>

#ifdef BME680
#include "Adafruit_BME680.h"
#define NumMenus 6
#else
#include "Adafruit_BME280.h"
#define NumMenus 5
#endif

#define BME_SCK 13
#define BME_MISO 12
#define BME_MOSI 11
#define BME_CS 10

#define MagicNumber 12345 // This is a 16 bit int, don't pick a number greater than 32K

#define SEALEVELPRESSURE_HPA (1013.25)

#ifdef BME680
Adafruit_BME680 bme; // I2C
//Adafruit_BME680 bme(BME_CS); // hardware SPI
//Adafruit_BME680 bme(BME_CS, BME_MOSI, BME_MISO,  BME_SCK);
#endif

#ifdef BME280
Adafruit_BME280 bme; // I2C
//Adafruit_BME280 bme(BME_CS); // hardware SPI
//Adafruit_BME280 bme(BME_CS, BME_MOSI, BME_MISO, BME_SCK); // software SPI
#endif

// For the breakout, you can use any 2 or 3 pins
// These pins will also work for the 1.8" TFT shield
#define TFT_CS     10
#define TFT_RST    9  // you can also connect this to the Arduino reset
                      // in which case, set this #define pin to 0!
#define TFT_DC     8

// Option 1 (recommended): must use the hardware SPI pins
// (for UNO thats sclk = 13 and sid = 11) and pin 10 must be
// an output. This is much faster - also required if you want
// to use the microSD card (see the image drawing example)
Adafruit_ST7735 tft = Adafruit_ST7735(TFT_CS,  TFT_DC, TFT_RST);

// Option 2: use any pins but a little slower!
//#define TFT_SCLK 13   // set these to be whatever pins you like!
//#define TFT_MOSI 11   // set these to be whatever pins you like!
//Adafruit_ST7735 tft = Adafruit_ST7735(TFT_CS, TFT_DC, TFT_MOSI, TFT_SCLK, TFT_RST);

// Joystick buttons on the Adfruit display
#define BUTTON_NONE 0
#define BUTTON_DOWN 1
#define BUTTON_RIGHT 2
#define BUTTON_SELECT 3
#define BUTTON_UP 4
#define BUTTON_LEFT 5

//use these constants to minimize calsulations at runtime (runs faster to precalculate)
const float c2f = 9.0/5.0;
const float hpainhg = 0.0001450377F * 2.03602;

//Global floating point numbers for temperature, pressure, humidity, altitude, and gas
float T_new,T_old,P_new,P_old,AL_new,AL_old,H_new,H_old;

#ifdef BME680
float G_new,G_old;
#endif

int TFTclean=0;

uint8_t buttonhistory = 0;

// calibration offsets
struct CalOffsets{
  const int magic_num=MagicNumber;
  float Toffset;
  float Hoffset;
  float Poffset;
  float Aoffset;
  #ifdef BME680  
  float Goffset;
  #endif
};

CalOffsets cal;

int SMS=0; // state machine state number
int SMSold=0; // old state machine state
int CalClean=0;

// this allows the code to print the old valuse in black. Effectively erasing the old test.
String OldType;
float  OldReading;
String OldUnits;
float  OldOffset;

//read joystick buttons
uint8_t readButton(void) {
  float a = analogRead(3);
  
  a *= 3.3; //5.0; connect aref to 3.3V, schematic shows 3.3V Ed, WA6ED
  a /= 1024.0;
  

  #ifdef DEBUG
  Serial.print("Button read analog = ");
  Serial.println(a);
  #endif

  // For this to work withthe joystick on battery, AREF needs to be connected to 3.3V
  if (a < 0.2) return BUTTON_LEFT;    //BUTTON_DOWN;
  if (a < 1.0) return BUTTON_DOWN;    //BUTTON_RIGHT;
  if (a < 1.5) return BUTTON_SELECT;  //
  if (a < 2.0) return BUTTON_RIGHT;   //BUTTON_UP;
  if (a < 3.2) return BUTTON_UP;      //BUTTON_LEFT;                  
  else return BUTTON_NONE;            // none
}

//print text that won't change only once. This saves update time later.
void PrintTemplate(){
  tft.fillScreen(ST7735_BLACK);
  tft.setTextColor(ST7735_GREEN);
  tft.setCursor(0,0);
  tft.println(F("T:   F"));
  tft.setCursor(0,20);
  tft.println(F("P:   inHg"));
  tft.setCursor(0,40);
  tft.println(F("H:   %"));
  tft.setCursor(0,60);
  tft.println(F("A:    Ft")); 
  #ifdef BME680
  tft.setCursor(0,80);
  tft.println(F("G:    KOhms"));
  #endif  
}

void setup(void) {

 analogReference(EXTERNAL);

 CalOffsets caltemp;
 EEPROM.get(0,caltemp);
 if(caltemp.magic_num == MagicNumber){
  EEPROM.get(0,cal);
 }else
 {
  cal.Toffset=0.0;
  cal.Hoffset=0.0;
  cal.Poffset=0.0;
  cal.Aoffset=0.0;
  #ifdef BME680
  cal.Goffset=0.0;
  #endif
 }

  #ifdef DEBUG
  Serial.begin(9600);
  #endif
  
  // Use this initializer if you're using a 1.8" TFT
  tft.initR(INITR_BLACKTAB);   // initialize a ST7735S chip, black tab
  // Use this initializer (uncomment) if you're using a 1.44" TFT
  //tft.initR(INITR_144GREENTAB);   // initialize a ST7735S chip, black tab

  tft.setRotation(3);
  tft.setTextSize(2);  
  tft.fillScreen(ST7735_BLACK);

  if (!bme.begin()) { //test if BME sensor is readable
    #ifdef BME680 
    tft.println(F("Could not find a valid BME680 sensor, check wiring!"));
    #else 
    tft.println(F("Could not find a valid BME280 sensor, check wiring!"));
    #endif
    while (1); // This is a wait forever
  }

  #ifdef BME680
  // Set up oversampling and filter initialization
  bme.setTemperatureOversampling(BME680_OS_8X);
  bme.setHumidityOversampling(BME680_OS_2X);
  bme.setPressureOversampling(BME680_OS_4X);
  bme.setIIRFilterSize(BME680_FILTER_SIZE_3);
  bme.setGasHeater(320, 150); // 320*C for 150 ms
  #endif
  
  tft.setTextSize(1); // Small text for splash screen
  tft.setTextColor(ST7735_YELLOW);  
  uint16_t time = millis();
  time = millis() - time;

  //
  // Notice this nice F("text") macro below? This is cool. It forces the compiler
  // to get the string from flash memory. If you don't use this macro at runtime the
  // string gets copied from flash memory to sram. And, the string is referenced
  // from sram to print. More of your sram is consumed with stuff that doesn't change.
  // Not good... So, this saves your valuable SRAM for variables that change.
  //
  #ifdef BME680
  tft.println(F("SOARA BME680&TFT Demo"));
  #else
  tft.println(F("SOARA BME280&TFT Demo"));
  #endif
  
  tft.println(F("Based on Adafruit"));
  tft.println(F("libraries & examples")); 
  tft.println(F("Revsion 0.37"));

  delay(1000); //wait 1 second before erasing splash screen and starting oprtation
  tft.setTextSize(2); // larger text for normal operation

  #ifdef BME680
  if (! bme.performReading()) {
    tft.setRotation(3);
    tft.setTextSize(1);  
    tft.fillScreen(ST7735_BLACK);
    tft.setTextColor(ST7735_RED);
    tft.setCursor(0,0);
    tft.println(F("Failed to perform reading :("));
    delay(1000);
    return;
  }
  T_new=(bme.temperature* c2f) + 32.0-3; // pre calculate values so delays fetching and crunching data will be transparent
  P_new=bme.pressure * hpainhg; 
  H_new=bme.humidity;
  G_new=bme.gas_resistance / 1000.0;
  AL_new=bme.readAltitude(SEALEVELPRESSURE_HPA)*3.28084;
  #else
  T_new=(bme.readTemperature()* c2f) + 32.0-3; // pre calculate values so delays fetching and crunching data will be transparent
  P_new=bme.readPressure() * hpainhg; 
  H_new=bme.readHumidity();
  AL_new=bme.readAltitude(SEALEVELPRESSURE_HPA)*3.28084;
  #endif
  
  PrintTemplate();     
}

void PrintTemp(int color){
 if(T_old != T_new){  
   tft.setCursor(22,0);
   tft.setTextColor(ST7735_BLACK); //print black characters to erase old text
   tft.print(T_old,0); //print last value in black to remove old text
   tft.setCursor(22,0);
   tft.setTextColor(color); //now print green with new data
   tft.print(T_new,0);
 }
}

void PrintPressure(int color){
  if(P_old != P_new){  
   tft.setCursor(22,20);
   tft.setTextColor(ST7735_BLACK);
   tft.print(P_old,0); //hPA * 0.01450377 = PSI or x 2.03602 inHg
   tft.setCursor(22,20);
   tft.setTextColor(color);
   tft.print(P_new,0); //hPA * 0.01450377 = PSI or x 2.03602 inHg
  }
}

void PrintHumidity(int color){
  if(H_old!=H_new){
   tft.setCursor(22,40);
   tft.setTextColor(ST7735_BLACK);
   tft.print(H_old,0); //print last value in black to remove old text
   tft.setCursor(22,40);
   tft.setTextColor(color); 
   tft.print(H_new,0);
  }
}

#ifdef BME680
void PrintGas(int color){
  if(G_old!=G_new){
   tft.setCursor(22,80);
   tft.setTextColor(ST7735_BLACK);
   tft.print(G_old,0);
   tft.setCursor(22,80);
   tft.setTextColor(color); 
   tft.print(G_new,0);
  }
}
#endif

void PrintAltitude(int color){
  if(AL_new != AL_old){  
   tft.setCursor(22,60);
   tft.setTextColor(ST7735_BLACK);
   tft.print(AL_old,0); //print last value in black to remove old text
   tft.setCursor(22,60);
   tft.setTextColor(color);
   tft.print(AL_new,0); //scale meters to feet
  } 
}

// We need to save the old values so we can print them later in BLACK.
// This will effectively erases the old text. Clearing the whole screen and rewriting everything
// is easier to code. But, rewriting everything is slower.
void SaveOldValues(){
  T_old=T_new;
  P_old=P_new;
  H_old=H_new;
  AL_old=AL_new;
  #ifdef BME680
  G_old=G_new;
  #endif
}

#ifdef BME680
void ReadBMEsensor(){
  // read data from sensor
  if (! bme.performReading()) {
    // if reading sensor data fails print an error to the screen
    tft.setRotation(3);
    tft.setTextSize(1);
    tft.fillScreen(ST7735_BLACK);
    tft.setTextColor(ST7735_RED);
    tft.setCursor(0,0);   
    tft.println(F("Failed to perform reading :("));
    delay(1000);    
    return;
  }
}
#endif

void CalculateNewReadings(){
   #ifdef BME680
   T_new=round((bme.temperature* c2f) + 32.0 + cal.Toffset);
   H_new=round(bme.humidity + cal.Hoffset);
   G_new=round((bme.gas_resistance / 1000.0) + cal.Goffset);
   P_new=round((bme.pressure * hpainhg) + cal.Poffset);
   AL_new=round((bme.readAltitude(SEALEVELPRESSURE_HPA)*3.28084) + cal.Aoffset);
   #endif

   #ifdef BME280
   T_new=round((bme.readTemperature()* c2f) + 32.0 + cal.Toffset);
   H_new=round(bme.readHumidity() + cal.Hoffset);
   P_new=round((bme.readPressure() * hpainhg) + cal.Poffset);
   AL_new=round((bme.readAltitude(SEALEVELPRESSURE_HPA)*3.28084) + cal.Aoffset);
   #endif
}

#ifdef DEBUG
void printstate(){
   tft.setCursor(148,0);
   tft.setTextColor(ST7735_BLACK); //print black characters to erase old text
   tft.print(abs(SMSold)); //print last value in black to remove old text
   tft.setCursor(148,0);
   tft.setTextColor(ST7735_RED);
   tft.print(abs(SMS));
}
#endif

void DecodeButton()
{
  uint8_t b = readButton();
  if (b == BUTTON_DOWN) {
     SMSold=SMS;
     SMS= --SMS % NumMenus;
     #ifdef DEBUG
     printstate();
     #endif
     
     while(readButton()==BUTTON_DOWN); //wait for button to be released
  }
  if (b == BUTTON_LEFT) {
     switch(abs(SMS)){
      case 1: cal.Toffset--; break;
      case 2: cal.Poffset--; break;
      case 3: cal.Hoffset--; break;
      case 4: cal.Aoffset--; break;
      #ifdef BME680
      case 5: cal.Goffset--; break;
      #endif
     }
     while(readButton()==BUTTON_LEFT); //wait for button to be released
  }
  if (b == BUTTON_UP) {
     SMSold=SMS;
     SMS= ++SMS % NumMenus;
    
     #ifdef DEBUG
     printstate();
     #endif
          
     while(readButton()==BUTTON_UP);
  }
  if (b == BUTTON_RIGHT) {
     switch(abs(SMS)){
      case 1: cal.Toffset++; break;
      case 2: cal.Poffset++; break;
      case 3: cal.Hoffset++; break;
      case 4: cal.Aoffset++; break;
      #ifdef BME680
      case 5: cal.Goffset++; break;
      #endif
     }
     while(readButton()==BUTTON_RIGHT); //wait for button to be released
  }
  if (b == BUTTON_SELECT) { 
     while(readButton()==BUTTON_SELECT); //wait for button to be released

     CalOffsets caltemp;

     // get the EEPROM values and test if current offsets are the same... or not
     EEPROM.get(0,caltemp);
     if((caltemp.magic_num == MagicNumber)&&
        (caltemp.Toffset == cal.Toffset)&&
        (caltemp.Poffset == cal.Poffset)&&
        (caltemp.Hoffset == cal.Hoffset)&&
        #ifdef BME680
        (caltemp.Goffset == cal.Goffset)&&
        #endif
        (caltemp.Aoffset == cal.Aoffset))
     {
      // don't save to EEPROM they already match
      // No point in writing to EEPROM if values match
      // This saves unnecessary write cycles
     }else{
      // Okay they are different, allow the update
      EEPROM.put(0,cal); // I protected this part of the code since writing EEPROM will wear the EEPROM
      tft.fillScreen(ST7735_RED);
      tft.setTextColor(ST7735_BLACK);
      tft.setCursor(0,0);
      tft.println("EEPROM Wr");
      tft.println("Press reset");
      tft.println("to continue");
      while(1); // Every time a EEPROM put is run, print a message and force a reset. This way you know you are not accidently writing to EEPROM unnecessarly
                // EEPROM will wear out if you write to it too much.
     }
    }
}

void PrintCalibrate(String type, float reading, String units, float offset){

  if(CalClean ==0){ 
   tft.fillScreen(ST7735_BLACK);
   tft.setTextColor(ST7735_RED);
   tft.setCursor(0,0);
   tft.println("Calibrate");
   tft.println("left  -1");
   tft.println("right +1");
   tft.println("pushin save");
   tft.println("");
   CalClean=1;
  }

   tft.setCursor(0,80);
   tft.setTextColor(ST7735_BLACK);     
   tft.print(OldType);
   tft.print(OldReading,0);
   tft.println(OldUnits); 
   tft.print(F("offset="));
   tft.println(OldOffset,0);

   tft.setTextColor(ST7735_RED);
   tft.setCursor(0,80);  
   tft.print(type);
   tft.print(reading,0);
   tft.println(units); 
   tft.print(F("offset="));
   tft.println(offset,0);

   OldType=type;
   OldReading=reading;
   OldUnits=units;
   OldOffset=offset;
}

void loop() {
   DecodeButton();
   
   #ifdef BME680
   ReadBMEsensor();
   #endif
   
    switch(abs(SMS)){
    case 0:
           {
            if(!TFTclean){
              tft.fillScreen(ST7735_BLACK);
              PrintTemplate();
              TFTclean=1;
            }
            PrintTemp(ST7735_GREEN);
            PrintPressure(ST7735_GREEN);
            PrintHumidity(ST7735_GREEN);
            PrintAltitude(ST7735_GREEN);
            #ifdef BME680
            PrintGas(ST7735_GREEN);
            #endif
            CalClean=0;            
            break;
           }
    case 1:
           {
             TFTclean=0;
             PrintCalibrate("T:",T_new,"F",cal.Toffset);
             break;
           }
    case 2:
           {
             TFTclean=0;
             PrintCalibrate("P:",P_new,"inHg",cal.Poffset);
             break;
           }
    case 3:
           {
             TFTclean=0;
             PrintCalibrate("H:",H_new,"%",cal.Hoffset);                        
             break;
           }

    case 4:
           {
             TFTclean=0;
             PrintCalibrate("A:",AL_new,"Ft",cal.Aoffset);                       
             break;
           }
    #ifdef BME680       
    case 5:
           {
             TFTclean=0;
             PrintCalibrate("G:",G_new,"KOhms",cal.Goffset);                        
             break;
           }
    #endif
   }

  SaveOldValues();
  CalculateNewReadings();
  
  #ifdef DEBUG
  printstate();
  #endif
 }

